#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import pickle
from datetime import datetime

from hypothesis import given
from hypothesis import strategies as st

from xotless.ranges import Range
from xotless.trees import Cell, IntervalTree

from xotless.testing.strategies.domains import (
    FloatRangeHelper,
    date_ranges,
    float_ranges,
    many_date_ranges,
    many_float_ranges,
    members_of_date_ranges,
    members_of_float_ranges,
)

MORE_THAN_ONE_LIFETIME = Range.new_open_right(datetime(1978, 10, 21), datetime(9078, 10, 21))


@given(st.data(), many_date_ranges(5, outer=MORE_THAN_ONE_LIFETIME))
def test_date_interval_trees_finds_the_right_interval(data, intervals):
    cells = [Cell.from_range(r, i) for i, r in enumerate(intervals)]
    tree = IntervalTree.from_cells(cells)
    picked_cell = data.draw(st.sampled_from(cells))
    interval = Range.new_open_right(picked_cell.lowerbound, picked_cell.upperbound)
    member = data.draw(members_of_date_ranges(interval))
    assert tree[member] == (picked_cell,)
    for protocol in range(pickle.DEFAULT_PROTOCOL, pickle.HIGHEST_PROTOCOL):
        tree = pickle.loads(pickle.dumps(tree, protocol))
        assert tree[member] == (picked_cell,)


# IMPORTANT: These values make the FLOAT_OUTER_RANGE completely contained
# inside FloatRangeHelper's range, and the both are far-enough so that we can
# get f64 values between in the ranges (FloatRangeHelper.LOWER_LIMIT,
# FLOAT_LOWER_LIMIT) and (FLOAT_UPPER_LIMIT, FloatRangeHelper.UPPER_LIMIT)
#
# This is import to keep cause members_of_float_ranges requires that any
# possible range has f64 values.  Otherwise you may get an InvalidArgument
# error like:
#
#  hypothesis.errors.InvalidArgument: There are no 64-bit floating-point
#  values between min_value=-10000000000.0 and max_value=-10000000001.0
#
FLOAT_LOWER_LIMIT = FloatRangeHelper.LOWER_LIMIT + 1000
FLOAT_UPPER_LIMIT = FloatRangeHelper.UPPER_LIMIT - 1000
FLOAT_OUTER_RANGE = Range.new_open_right(FLOAT_LOWER_LIMIT, FLOAT_UPPER_LIMIT)


@given(st.data(), many_float_ranges(5, outer=FLOAT_OUTER_RANGE))
def test_float_interval_trees_finds_the_right_interval(data, intervals):
    cells = [Cell.from_range(r, i) for i, r in enumerate(intervals)]
    tree = IntervalTree.from_cells(cells)
    picked_cell = data.draw(st.sampled_from(cells))
    interval = Range.new_open_right(picked_cell.lowerbound, picked_cell.upperbound)
    member = data.draw(members_of_float_ranges(interval))
    assert tree[member] == (picked_cell,)
    for protocol in range(pickle.DEFAULT_PROTOCOL, pickle.HIGHEST_PROTOCOL):
        tree = pickle.loads(pickle.dumps(tree, protocol))
        assert tree[member] == (picked_cell,)


@given(st.data(), many_float_ranges(5, outer=FLOAT_OUTER_RANGE))
def test_float_interval_trees_finds_the_right_interval_issue3_regression(data, intervals):
    cells = [Cell.from_range(r, i) for i, r in enumerate(intervals)]
    # The following two cells are regression tests for issue
    # https://gitlab.merchise.org/mercurio-2018/xotless/-/issues/3
    cells.append(Cell.from_bounds(None, FLOAT_LOWER_LIMIT - 1, len(cells)))
    cells.append(Cell.from_bounds(FLOAT_UPPER_LIMIT + 1, None, len(cells)))
    tree = IntervalTree.from_cells(cells)
    picked_cell = data.draw(st.sampled_from(cells))
    interval = Range.new_open_right(picked_cell.lowerbound, picked_cell.upperbound)
    member = data.draw(members_of_float_ranges(interval))
    assert tree[member] == (picked_cell,)
    for protocol in range(pickle.DEFAULT_PROTOCOL, pickle.HIGHEST_PROTOCOL):
        tree = pickle.loads(pickle.dumps(tree, protocol))
        assert tree[member] == (picked_cell,)


@given(
    st.data(),
    st.lists(
        float_ranges(
            outer=Range.new_open_right(-1e10, 1e10),
            kinds=st.just(Range.new_open_right),
        ),
        min_size=1,
        max_size=10,
    ),
)
def test_float_interval_trees_finds_only_the_right_intervals(data, intervals):
    cells = [Cell.from_range(r, i) for i, r in enumerate(intervals)]
    tree = IntervalTree.from_cells(cells)
    picked_cell = data.draw(st.sampled_from(cells))
    interval = Range.new_open_right(picked_cell.lowerbound, picked_cell.upperbound)
    member = data.draw(members_of_float_ranges(interval))
    expected = {cell for cell in cells if member in cell}
    assert set(tree[member]) == expected
    if len(cells):
        assert len(expected)
    for protocol in range(pickle.DEFAULT_PROTOCOL, pickle.HIGHEST_PROTOCOL):
        tree = pickle.loads(pickle.dumps(tree, protocol))
        assert set(tree[member]) == expected


@given(
    st.data(),
    st.lists(
        date_ranges(outer=MORE_THAN_ONE_LIFETIME, kinds=st.just(Range.new_open_right)),
        min_size=1,
        max_size=10,
    ),
)
def test_date_interval_trees_finds_only_the_right_intervals(data, intervals):
    cells = [Cell.from_range(r, i) for i, r in enumerate(intervals)]
    tree = IntervalTree.from_cells(cells)
    picked_cell = data.draw(st.sampled_from(cells))
    interval = Range.new_open_right(picked_cell.lowerbound, picked_cell.upperbound)
    member = data.draw(members_of_date_ranges(interval))
    expected = {cell for cell in cells if member in cell}
    assert set(tree[member]) == expected
    if len(cells):
        assert len(expected)
    for protocol in range(pickle.DEFAULT_PROTOCOL, pickle.HIGHEST_PROTOCOL):
        tree = pickle.loads(pickle.dumps(tree, protocol))
        assert set(tree[member]) == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
"""Implementation of several domains.

Domains are used by AVMs to describe groups of values of an attribute for
which a program behaves in the same way.

"""

from collections import deque
from dataclasses import dataclass
from itertools import product
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from xotl.tools.objects import lazy, setdefaultattr

from .ranges import Range
from .tracing import get_module_sentry_spanner
from .types import Domain, TEq, TOrd


@dataclass(init=False, unsafe_hash=True)
class EquivalenceSet(Generic[TEq]):
    "A simple wrapper over a Python set for items with EqTypeClass"

    values: AbstractSet[TEq]
    __slots__ = ("values", "_sample")

    def __init__(self, values: AbstractSet[TEq]) -> None:
        self.values = frozenset(values)

    def to_interval_set(self) -> "IntervalSet[TOrd]":
        """Convert to a semantically-equivalent interval set.

        This should only be applied for equivalence sets containing orderable
        (Ord) items.

        """
        ranges: List[Range[TOrd]] = []
        values = set(self.values)
        while values:
            val: TOrd = values.pop()  # type: ignore
            ranges.append(Range.new_closed(val, val))
        return IntervalSet(ranges)

    def __bool__(self):
        return bool(self.values)

    def __contains__(self, val: TEq) -> bool:
        return val in self.values

    def __le__(self, other: "EquivalenceSet") -> bool:
        if isinstance(other, IntervalSet):
            return self.to_interval_set() <= other
        if not isinstance(other, EquivalenceSet):
            return NotImplemented
        return self.values <= other.values

    def __lt__(self, other: "EquivalenceSet") -> bool:  # pragma: no cover
        if isinstance(other, IntervalSet):
            return self.to_interval_set() < other
        if not isinstance(other, EquivalenceSet):
            return NotImplemented
        return self.values < other.values

    def __ge__(self, other: "EquivalenceSet") -> bool:  # pragma: no cover
        if isinstance(other, IntervalSet):
            return self.to_interval_set() >= other
        if not isinstance(other, EquivalenceSet):
            return NotImplemented
        return self.values >= other.values

    def __gt__(self, other: "EquivalenceSet") -> bool:  # pragma: no cover
        if isinstance(other, IntervalSet):
            return self.to_interval_set() > other
        if not isinstance(other, EquivalenceSet):
            return NotImplemented
        return self.values > other.values

    def __and__(
        self,
        other: "EquivalenceSet",
    ) -> "EquivalenceSet":  # pragma: no cover
        if isinstance(other, IntervalSet):
            return self.to_interval_set() & other
        if not isinstance(other, EquivalenceSet):
            return NotImplemented
        return EquivalenceSet(self.values & other.values)

    def diff(
        self,
        other: "EquivalenceSet",
    ) -> "EquivalenceSet":  # pragma: no cover
        if isinstance(other, IntervalSet):
            return self.to_interval_set().diff(other)
        if not isinstance(other, EquivalenceSet):
            return NotImplemented
        return EquivalenceSet(self.values - other.values)

    __sub__ = difference = diff

    def __or__(
        self,
        other: "EquivalenceSet",
    ) -> "EquivalenceSet":  # pragma: no cover
        if isinstance(other, IntervalSet):
            return self.to_interval_set() | other
        if not isinstance(other, EquivalenceSet):
            return NotImplemented
        return EquivalenceSet(self.values | other.values)

    def union(
        self,
        *others: "EquivalenceSet",
    ) -> "EquivalenceSet":  # pragma: no cover
        from functools import reduce
        from operator import or_

        sets: Tuple[EquivalenceSet, ...] = tuple(s for s in (self,) + others if s)
        values: AbstractSet = reduce(or_, (s.values for s in sets), frozenset())
        return EquivalenceSet(values)

    join = union

    @property
    def sample(self) -> Optional[TEq]:
        if self.values:
            # We cache the result to return the same sample all the time.
            # set.pop() return an arbitrary element.
            return setdefaultattr(self, "_sample", lazy(lambda: set(self.values).pop()))
        else:
            return None


class DomainMismatch(TypeError):
    pass


@dataclass(init=False, unsafe_hash=True)
class IntervalSet(Generic[TOrd]):
    ranges: Tuple[Range[TOrd], ...]
    __slots__ = ("ranges",)

    def __init__(self, bases: Iterable[Range[TOrd]]) -> None:
        if bases:
            # Normalize the bases, we must keep our individual bases
            # disjoint.
            base, *others = self._extract_all_ranges(bases)
            # Ensure the first base range is not empty
            while not base and others:
                base, *others = others
            if base:
                normalized = base.join(*others)
                assert all(r for r in normalized)
                # `join` will return non-empty bases since `base` is non-empty
                self.ranges = normalized
            else:
                assert not others
                self.ranges = ()
        else:
            self.ranges = ()

    @classmethod
    def _from_ordered_disjoint(
        cls,
        bases: Iterable[Range[TOrd]],
    ) -> "IntervalSet[TOrd]":
        result = cls([])
        result.ranges = tuple(bases)
        return result

    @classmethod
    def _extract_ranges(
        cls,
        other: Union["IntervalSet[TOrd]", Range[TOrd]],
    ) -> Tuple[Range[TOrd], ...]:
        if isinstance(other, EquivalenceSet):
            return other.to_interval_set().ranges
        if isinstance(other, IntervalSet):
            return other.ranges
        else:
            assert isinstance(other, Range)
            return (other,)

    @classmethod
    def _extract_all_ranges(
        cls,
        *args: Iterable[Union["IntervalSet[TOrd]", Range[TOrd]]],
    ) -> Tuple[Range[TOrd], ...]:
        return tuple(base for arg in args for bases in arg for base in cls._extract_ranges(bases))

    def __repr__(self):
        return f"IntervalSet({self.ranges!r})"

    @property
    def sample(self) -> Optional[TOrd]:
        if self.ranges:
            return next((r.sample for r in self.ranges if r.sample is not None), None)
        else:
            return None

    def __contains__(self, value: TOrd) -> bool:
        return any(value in base for base in self.ranges)

    def __bool__(self):
        return bool(self.ranges)

    def __le__(self, other: "IntervalSet[TOrd]") -> bool:
        if isinstance(other, EquivalenceSet):
            other = other.to_interval_set()
        if not isinstance(other, IntervalSet):
            return NotImplemented
        if not self:
            return True
        if not other:
            return False
        return all(any(mine <= their for their in other.ranges) for mine in self.ranges)

    def __lt__(self, other: "IntervalSet[TOrd]") -> bool:  # pragma: no cover
        if isinstance(other, EquivalenceSet):
            other = other.to_interval_set()
        if not isinstance(other, IntervalSet):
            return NotImplemented
        return self <= other and not (other <= self)

    def __ge__(self, other: "IntervalSet[TOrd]") -> bool:  # pragma: no cover
        if isinstance(other, EquivalenceSet):
            other = other.to_interval_set()
        if not isinstance(other, IntervalSet):
            return NotImplemented
        return other <= self

    def __gt__(self, other: "IntervalSet[TOrd]") -> bool:  # pragma: no cover
        if isinstance(other, EquivalenceSet):
            other = other.to_interval_set()
        if not isinstance(other, IntervalSet):
            return NotImplemented
        return self >= other and not (other >= self)

    def __eq__(self, other) -> bool:
        if isinstance(other, EquivalenceSet):
            other = other.to_interval_set()
        if not isinstance(other, IntervalSet):
            return NotImplemented
        return self <= other <= self

    def __and__(self, other: "IntervalSet[TOrd]") -> "IntervalSet[TOrd]":
        if isinstance(other, EquivalenceSet):
            other = other.to_interval_set()
        if not isinstance(other, IntervalSet):
            return NotImplemented
        bases = tuple(
            filter(
                None,
                (base & other_base for base, other_base in product(self.ranges, other.ranges)),
            )
        )
        return IntervalSet(bases)

    def union(self, *others: "IntervalSet[TOrd]") -> "IntervalSet[TOrd]":
        with sentry_span("IntervalSet.union"):
            all_ranges = self._extract_all_ranges(self.ranges, others)
            if all_ranges:
                base, *bases = all_ranges
                new_bases = base.join(*bases)
            else:
                new_bases = ()
            return IntervalSet(new_bases)

    __or__ = join = union

    def diff(self, other: "IntervalSet[TOrd]") -> "IntervalSet[TOrd]":
        # The algorithm depends that 'ranges' are sorted ranges from left to
        # right by the lowerbound, like this:
        #
        #    ours      |------|   |------|   |--| |-----| |---|
        #  theirs   |---------------|  |-|          |-------|
        #
        # Our __init__ ensures this.
        with sentry_span("IntervalSet.diff"):
            if isinstance(other, EquivalenceSet):
                other = other.to_interval_set()
            ours = deque(self.ranges)
            theirs = deque(other.ranges)
            results = []
            while ours and theirs:
                top: Range = ours.popleft()
                bottom: Range = theirs.popleft()
                assert top and bottom
                if top == bottom:
                    # Nothing to produce here, both sub-ranges are consumed.
                    pass
                elif top & bottom:
                    # There are four possible cases here (five if we count top
                    # == bottom, but we ruled that out):
                    #
                    # case:       a     !     b    !    c   !    d
                    # ~~~~~~~~~~~~~~~~~~+~~~~~~~~~~+~~~~~~~~+~~~~~~~~~
                    # top      |------| !   |----| ! |---|  !   |--|
                    # bottom     |--|   ! |---|    !   |---|! |------|
                    #
                    # The case 'a' will produce a range to "left" that cannot
                    # intercept with any other range; so we can put it the
                    # results.  The second range may intercept with another of
                    # theirs so we put it in the front of ours.  But this
                    # case, has two sub-cases:
                    #
                    #                a.1               a.2
                    #   top        |-----|   and    |-----|
                    #   bottom     |---|               |--|
                    #
                    # In the first sub-case we cannot push the diff to the
                    # result.  In the second we have nothing to push to ours.
                    # But those can be treated the same as the cases 'b' and
                    # 'c'.
                    diff = top.diff(bottom)
                    if diff:
                        if len(diff) > 1:  # ->                                   case 'a'.
                            left, right = diff
                            results.append(left)
                            ours.appendleft(right)
                        else:
                            assert len(diff) == 1
                            if (
                                bottom.lowerbound <= top.lowerbound
                            ):  # ->                        cases 'b' or a.1
                                ours.appendleft(diff[0])
                            elif (
                                top.upperbound <= bottom.upperbound
                            ):  # ->                        cases 'c' or a.2
                                results.append(diff[0])
                                # We need to reintroduce bottom to theirs, because
                                # it can overlap with the next top of 'ours'.
                                theirs.appendleft(bottom)
                            else:
                                assert False
                    else:  # ->                                  case 'd'
                        # We need to reintroduce bottom to theirs, because it
                        # can overlap with the next top of 'ours'.
                        theirs.appendleft(bottom)
                else:
                    # There are two cases here:
                    #
                    #                 'a'               'b'
                    #   top             |-----|   and    |-----|
                    #   bottom     |---|                        |--|
                    #
                    if bottom.upperbound <= top.lowerbound:
                        ours.appendleft(top)
                    elif top.upperbound <= bottom.lowerbound:
                        theirs.appendleft(bottom)
                    else:  # pragma: no cover
                        assert False

            if ours:
                assert not theirs
                results.extend(ours)
            return IntervalSet._from_ordered_disjoint(results)

    __sub__ = difference = diff


def iter_domain_endpoints(domains: Sequence[Domain[TOrd]]) -> Iterator[TOrd]:
    """Yields likely endpoints in the domains.

    We don't actually yield all the endpoints but those who can be one of the
    max/min endpoint of an entire interval which covers all the domains.

    Basically, for each domain we yield at most two endpoints: the min and max
    of the endpoints in the domain.  For an
    `~xotless.domains.EquivalenceSet`:class:, the endpoints are just the
    min/max values in the set.  If the set has a single value, we only yield
    once.  For an `~xotless.domains.IntervalSet`:class: the endpoints are the
    min and max all the ranges regardless of if they are included or not in
    the interval set.

    Example:

       >>> eq1 = EquivalenceSet({1, -1, 4})
       >>> iset1 = IntervalSet([Range.new_open_right(0, 1), Range.new_open_right(1, 2)])
       >>> list(iter_domain_endpoints([eq1, iset1]))
       [-1, 4, 0, 2]

    """
    for domain in domains:
        if isinstance(domain, EquivalenceSet):
            yield min(domain.values)
            if len(domain.values) > 1:
                yield max(domain.values)
        elif isinstance(domain, IntervalSet):
            # IntervalSet keep their ranges sorted by the lower bound, so the
            # first has the minimal value.
            yield domain.ranges[0].lowerbound
            yield max(r.upperbound for r in domain.ranges)
        else:
            assert False, f"Unknown type of domain {domain.__class__}"


sentry_span = get_module_sentry_spanner(__name__)


if TYPE_CHECKING:  # pragma: no cover
    from datetime import datetime, timedelta

    def check(domain: Domain) -> None:
        return None

    def check_subdomain(domain: Domain[datetime]) -> None:
        return None

    check(EquivalenceSet({1}))
    check(Range.new_open_right(datetime.now(), datetime.now() + timedelta(1)).lift())
    check_subdomain(Range.new_open_right(datetime.now(), datetime.now() + timedelta(1)).lift())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from collections import deque
from dataclasses import dataclass
from typing import Iterator

from typing_extensions import Protocol
from xotl.tools.string import cut_prefix, cut_suffix


# fmt: off
class Node(Protocol):
    def __hash__(self): ...
    def __iter__(self) -> Iterator["Node"]: ...
# fmt: on


@dataclass(unsafe_hash=True)
class NodeRef:
    "A wrapper around a node in walks."

    node: Node

    def __repr__(self):
        node = cut_suffix(cut_prefix(repr(self.node), "<"), ">")
        return f"<&{node}>"


def pruned_prefix_walk(
    node: Node,
    prune_entirely: bool = False,
) -> Iterator[NodeRef]:
    r"""Produced a pruned prefix walk of a Directed Acyclic Graph.

    By default, pruned means that the same subgraph is never yielded more than
    once.  However to provide the full information for every node in the graph
    we yield references to nodes.  This allows to *repeat* nodes without yield
    the whole subgraph. If `pruned_entirely` is True, never yield the same
    node more than once.

    Example:

    Let's make a graph like this::

        a ------> c -----------------.
        |        / \                 |
        | .-----'   '----\           v
        | |               '-> f <--- e
        v v                   |
         b  -----> d <--------'

    First, just boilerplate to have a Item which has the Node interface:

        >>> from dataclasses import dataclass
        >>> from typing import Tuple

        >>> @dataclass(unsafe_hash=True)
        ... class Item:
        ...     name: str
        ...     children: Tuple["Item", ...]
        ...     def __iter__(self):
        ...         return iter(self.children)
        ...     def __repr__(self):
        ...         return f"<{self.name}>"

    Now, let's build the graph:

        >>> d = Item("d", ())
        >>> f = Item("f", (d, ))
        >>> e = Item("e", (f, ))
        >>> b = Item("b", (d, ))
        >>> c = Item("c", (b, f, e))
        >>> a = Item("a", (b, c))

    And finally, let's walk the graph:

        >>> from xotless.walk import pruned_prefix_walk
        >>> list(pruned_prefix_walk(a))
        [<&a>, <&b>, <&d>, <&c>, <&b>, <&f>, <&d>, <&e>, <&f>]

    Notice that we reach node 'b' a second time when coming from node 'c', yet
    this time we don't follow 'b' and that's why we don't yield 'd'
    immediately after the second yield of 'b'.  Nevertheless, node 'd' does
    appear a second time, but because reach it from node 'f'.

    If `prune_entirely` is set to True we don't get any duplicate:

        >>> list(pruned_prefix_walk(a, prune_entirely=True))
        [<&a>, <&b>, <&d>, <&c>, <&f>, <&e>]


    """
    memory = set([])
    nodes = deque([node])
    while nodes:
        current = nodes.popleft()
        yield NodeRef(current)
        if current not in memory:
            memory.add(current)
            children = [n for n in iter(current) if not prune_entirely or n not in memory]
            children.reverse()
            nodes.extendleft(children)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import contextlib
from functools import partial

try:
    from sentry_sdk import Hub

    def sentry_span(module, description, **tags):
        hub = Hub.current
        span = hub.start_span(op=module, description=description)
        for tag, value in tags.items():
            span.set_tag(tag, value)
        return span

except ImportError:

    @contextlib.contextmanager
    def sentry_span(module, description, **tags):
        yield


def get_module_sentry_spanner(module):
    """Return a function to create Sentry spans for `module`.

    Usage:

       >>> sentry_span = get_module_sentry_spanner(__name__)       # doctest: +SKIP
       >>> with sentry_span("function_name_or_whatever", **tags):  # doctest: +SKIP
       ...    pass

    """
    return partial(sentry_span, module)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import operator
import pickle
import unittest
from copy import copy
from datetime import timedelta
from itertools import combinations
from random import sample, shuffle

from hypothesis import HealthCheck, example, given, settings
from hypothesis import strategies as st
from hypothesis.strategies import SearchStrategy
from xotl.tools.infinity import Infinity

from xotless.domains import EquivalenceSet, IntervalSet
from xotless.ranges import Bound, Excluded, Included, Range, RangeKind

from xotless.testing.strategies.domains import (
    date_ranges,
    equivalence_sets,
    float_ranges,
    get_date_member_strategy,
    many_date_ranges,
    many_float_ranges,
    members_of_date_ranges,
    members_of_float_ranges,
    non_empty_equivalence_sets,
    type_of_bounds,
)

DAY = timedelta(1)


kinds_with_samples = st.sampled_from((Range.new_open_left, Range.new_open_right, Range.new_closed))


@st.composite
def _sorted_points(draw, many=2):
    assert many >= 2
    points = list(sorted(draw(st.sets(st.integers(), min_size=many, max_size=many))))
    if draw(st.booleans()):
        points[0] = -Infinity
    if draw(st.booleans()):
        points[-1] = Infinity
    return tuple(points)


ranges = st.tuples(
    # Generates Ranges by applying zip_map to a list of two Bound
    # to a sorted pair of integers.
    st.lists(type_of_bounds, min_size=2, max_size=2),
    _sorted_points(),
).map(lambda args: Range(*zip_map(args[0], args[1])))

integer_intervalsets: SearchStrategy[IntervalSet[int]] = st.sets(ranges).map(
    lambda s: IntervalSet(s)
)
integer_nonempty_intervalsets: SearchStrategy[IntervalSet[int]] = st.sets(
    ranges.filter(bool), min_size=1
).map(lambda s: IntervalSet(s))


@given(
    st.sets(st.integers(), min_size=3, max_size=3),
    type_of_bounds,
    type_of_bounds,
)
def test_range_intersection_with_a_common_boundary(points, Bound1, Bound2):
    a, b, c = sorted(points)
    r1 = Range(Bound1(a), Included(b))
    r2 = Range(Included(b), Bound2(c))
    assert r1 & r2 == Range(Included(b), Included(b))


@given(
    st.sets(st.integers(), min_size=3, max_size=3),
    type_of_bounds,
    type_of_bounds,
)
def test_range_intersection_with_a_common_Excluded_boundary(points, Bound1, Bound2):
    a, b, c = sorted(points)
    assert not Range(Bound1(a), Excluded(b)) & Range(Included(b), Bound2(c))
    assert not Range(Bound1(a), Included(b)) & Range(Excluded(b), Bound2(c))
    assert not Range(Bound1(a), Excluded(b)) & Range(Excluded(b), Bound2(c))


@given(
    st.sets(st.integers(), min_size=4, max_size=4),
    st.lists(type_of_bounds, min_size=4, max_size=4),
)
def test_range_intersection_with_no_common_bounds(points, Bounds):
    a, b, c, d = zip_map(Bounds, sorted(points))
    assert not Range(a, b) & Range(c, d)


@given(
    st.sets(st.integers(), min_size=4, max_size=4),
    st.lists(type_of_bounds, min_size=4, max_size=4),
)
def test_range_intersection_with_overlapping_bounds(points, Bounds):
    a, b, c, d = zip_map(Bounds, sorted(points))
    assert Range(a, c) & Range(b, d) == Range(b, c)


@given(ranges)
def test_range_intersection_with_overlapping_bounds_and_unbounded(r0):
    r1 = Range(None, r0._upper)
    r2 = Range(r0._lower, None)
    assert r1 & r2 == r0


@given(ranges, ranges)
def test_range_intersection_is_contained_in_the_original_ranges(r1, r2):
    assert r1 & r2 <= r1
    assert r1 & r2 <= r2


@given(ranges, ranges)
def test_range_intersection_is_commutative(r1, r2):
    assert r1 & r2 == r2 & r1


@given(ranges, ranges, ranges)
def test_range_intersection_is_associative(r1, r2, r3):
    assert (r1 & r2) & r3 == r1 & (r2 & r3)


@given(
    st.sets(st.integers(), min_size=4, max_size=4),
    st.lists(type_of_bounds, min_size=4, max_size=4),
)
def test_range_diff_totally_enclosed_range(points, Bounds):
    a, b, c, d = zip_map(Bounds, sorted(points))
    big = Range(a, d)
    small = Range(b, c)
    # So we have something like:
    #
    #   big:   a ------------------------ d
    # small:            b ------- c
    #
    # So, big.diff(small) will result in two ranges:
    #
    #  before:  a ---- ~b
    #  after:                     ~c ---- d
    #
    # ~b means that if b is Included in `small`, it must be Excluded in
    # `before`; if it was Excluded in `small`, it must be Included in
    # `before`.
    result = big.diff(small)
    before, after = result
    assert before == Range(a, ~b)
    assert after == Range(~c, d)


@given(ranges, ranges)
@example(Range.new_open_right(0, None), Range(Excluded(0), Included(1)))
def test_range_diff_is_contained_in_r1_and_missing_in_r2(r1, r2):
    for diff in r1.diff(r2):
        assert diff <= r1
        assert not (diff <= r2)


@given(
    st.sets(st.integers(), min_size=4, max_size=4),
    st.lists(type_of_bounds, min_size=4, max_size=4),
)
def test_range_diff_disjoint_range(points, Bounds):
    a, b, c, d = zip_map(Bounds, sorted(points))
    r1 = Range(a, b)
    r2 = Range(c, d)
    assert r1.diff(r2) == (r1,)
    assert r2.diff(r1) == (r2,)


@given(
    st.sets(st.integers(), min_size=4, max_size=4),
    st.lists(type_of_bounds, min_size=4, max_size=4),
)
def test_range_diff_overlapping_ranges(points, Bounds):
    a, b, c, d = zip_map(Bounds, sorted(points))
    r1 = Range(a, c)
    r2 = Range(b, d)
    (before,) = r1.diff(r2)
    assert before == Range(a, ~b)
    (after,) = r2.diff(r1)
    assert after == Range(~c, d)


@given(_sorted_points(many=3), type_of_bounds, type_of_bounds)
def test_range_join_doesnt_get_confused_about_excluded(points, Bound1, Bound2):
    a, b, c = points
    r1 = Range(Bound1(a), Excluded(b))
    r2 = Range(Excluded(b), Bound2(c))
    assert r1.join(r2) == (r1, r2)


@given(_sorted_points(many=3), type_of_bounds, type_of_bounds)
def test_range_join_collapse_included(points, Bound1, Bound2):
    a, b, c = points
    r1 = Range(Bound1(a), Included(b))
    r2 = Range(Excluded(b), Bound2(c))
    assert r1.join(r2) == (Range(Bound1(a), Bound2(c)),)


@given(_sorted_points(many=5), type_of_bounds, type_of_bounds)
def test_range_multi_bounded_joins(points, Bound1, Bound2):
    # So we have something like:
    #  r1  a ------------------------------------ e
    #  r2                        !=d ------------ e
    #  r3     b -------- ~c
    #  r4                 c -----  d
    #  r5                ~c ----- ~d
    #
    # So now:
    #  -> any range join with r1 will result in r1
    #  -> any r2 and r3 will not overlap so it will result in (r3, r2)
    #  -> any join of (r3,r4,r2) return Range(b, e)
    #  -> any join of (r3,r5,r2) return in the same (r3,r5,r2)
    #
    #  ~c and ~d means that this bounds are Excluded in r2,r3,r5 and
    # Included the other Ranges

    a, b, c, d, e = points
    r1 = Range(Bound1(a), Bound2(e))
    r2 = Range(Excluded(d), Bound2(e))
    r3 = Range(Bound1(b), Excluded(c))
    r4 = Range(Included(c), Included(d))
    r5 = Range(Excluded(c), Excluded(d))
    assert r1.join(r2, r3, r4) == (r1,)
    assert r2.join(r3) == (r3, r2)
    assert r3.join(r4, r2) == (Range(Bound1(b), Bound2(e)),)
    assert r3.join(r5, r2) == (r3, r5, r2)


def test_regression_hash_compatible_with_eq_in_ranges():
    r1 = Range.new_open_right(0, 0)
    r2 = Range.new_open_right(1, 1)
    assert r1 == r2
    assert hash(r1) == hash(r2)


def _test_float_ranges_generation(data, r0, kinds=None):
    r1 = data.draw(float_ranges(outer=r0, kinds=kinds))
    assert r1 <= r0, f"{r1} is not a sub-range of {r0}"
    r2 = data.draw(float_ranges(inner=r1, kinds=kinds))
    assert r1 <= r2, f"{r1} is not a sub-range of {r2}"


@given(st.data(), float_ranges())
def test_float_ranges_generation(data, r0):
    _test_float_ranges_generation(data, r0)


@given(st.data())
def test_float_ranges_generation_regression_job_117109(data):
    _test_float_ranges_generation(data, Range.new_open(0, 1), kinds=st.just(Range.new_closed))


@given(st.data(), date_ranges())
def test_date_ranges_confined_generation(data, r0):
    r1 = data.draw(date_ranges(outer=r0))
    assert r1 <= r0, f"{r1} is not a sub-range of {r0}"
    r2 = data.draw(date_ranges(inner=r1))
    assert r1 <= r2, f"{r1} is not a sub-range of {r2}"


class DomainCase(unittest.TestCase):  # pragma: no cover
    def assertPickable(self, what):
        for proto in range(pickle.DEFAULT_PROTOCOL, pickle.HIGHEST_PROTOCOL):
            serialized = pickle.dumps(what, proto)
            self.assertEqual(what, pickle.loads(serialized))

    def assertEmpty(self, x, msg=None):
        if x:
            raise self.failureException(msg or f"{x} is not empty")

    def assertNonEmpty(self, x, msg=None):
        if not x:
            raise self.failureException(msg or f"{x} is empty")

    def assertOverlap(self, a, b, msg=None):
        r = a & b
        if not r:
            raise self.failureException(msg or f"{a} and {b} don't overlap")

    def assertDisjoint(self, a, b, msg=None):
        r = a & b
        if r:
            raise self.failureException(msg or f"{a} and {b} are not disjoint")

    def assertSampleIsMember(self, domain1, domain2):
        if domain1:
            if domain1.sample is not None:
                self.assertIn(domain1.sample, domain1 | domain2)
            else:
                self.assertIsInstance(domain1, IntervalSet)
                for r in domain1.ranges:
                    self.assertEqual(r.kind, RangeKind.OPEN)


class TestRange(DomainCase):
    @staticmethod
    def add(left, right):  # pragma: no cover
        if left and right:
            assert right.lowerbound in left or right.lowerbound == left.upperbound
            return Range.new_open_right(left.lowerbound, max(left.upperbound, right.upperbound))
        elif left:
            return left
        else:
            return right

    @given(many_float_ranges(n=10) | many_date_ranges(n=10))
    def test_generated_siblings_are_consecutive(self, siblings):
        last = None
        for right in siblings:
            self.add(last, right)
            last = right
        self.assertIsNot(last, None)
        self.assertEqual(len(siblings), 10)

    @given(many_float_ranges(n=10))
    def test_ranges_generates_disjoint_numbers(self, siblings):
        for left, right in combinations(siblings, 2):
            self.assertEmpty(
                left & right,
                f"{left!r} & {right!r} = {left & right}; and is not empty",
            )

    @given(many_date_ranges(n=10))
    def test_ranges_generates_disjoint_dates(self, siblings):
        for left, right in combinations(siblings, 2):
            self.assertEmpty(
                left & right,
                f"{left!r} & {right!r} = {left & right}; and is not empty",
            )

    @given(many_float_ranges(n=10) | many_date_ranges(n=10))
    def test_ranges_generates_orderered(self, siblings):
        lowers = [r.lowerbound for r in siblings]
        self.assertEqual(lowers, list(sorted(lowers)))

    @given(many_float_ranges(n=4) | many_date_ranges(n=4))
    def test_non_empty_siblings(self, siblings):
        for range_ in siblings:
            self.assertNonEmpty(range_)

    @given(many_float_ranges(n=4) | many_date_ranges(n=4))
    def test_join_is_disjoint(self, siblings):
        joined = Range.join(*siblings)
        for a, b in combinations(joined, 2):
            self.assertDisjoint(a, b)

    @given(many_float_ranges(n=4) | many_date_ranges(n=4))
    def test_join_is_not_order_sensitive(self, siblings):
        expected = Range.join(*siblings)
        shuffled = sample(siblings, len(siblings))
        joined = Range.join(*shuffled)
        self.assertEqual(joined, expected)

    @given(many_float_ranges(n=4) | many_date_ranges(n=4))
    def test_join_doesnt_mutate_args(self, siblings):
        copies = tuple(copy(r) for r in siblings)
        Range.join(*siblings)
        self.assertEqual(siblings, copies)

    def test_join_returns_the_empty_tuple(self):
        ranges = [Range.new_open_right(x, x) for x in range(5)]
        joined = Range.join(*ranges)
        self.assertEmpty(joined)

    @given(float_ranges() | date_ranges())
    def test_diff_with_self_returns_nothing(self, range_):
        empty = range_.diff(range_)
        self.assertEmpty(empty)

    @given(float_ranges() | date_ranges())
    def test_range_iterable(self, range_):
        low, up = range_
        self.assertEqual(range_._lower.bound, low)
        self.assertEqual(range_._upper.bound, up)

    @given(float_ranges() | date_ranges())
    def test_range_pickable(self, range_):
        self.assertPickable(range_)

    @given(
        float_ranges(kinds=st.just(Range.new_open_right))
        .filter(bool)
        .flatmap(lambda r: st.tuples(st.just(r), members_of_float_ranges(r)))
        | date_ranges(kinds=st.just(Range.new_open_right))
        .filter(bool)
        .flatmap(lambda r: st.tuples(st.just(r), members_of_date_ranges(r)))
    )
    def test_members_of_ranges(self, args):
        range_, member = args
        self.assertIn(member, range_)

    @given(float_ranges(kinds=kinds_with_samples) | date_ranges(kinds=kinds_with_samples))
    def test_sample_in_range(self, range_):
        self.assertIn(range_.sample, range_, f"Sample {range_.sample} is not in {range_}")

    @given(float_ranges() | date_ranges())
    def test_range_kinds(self, range_):
        if isinstance(range_._lower, Included) and isinstance(range_._upper, Excluded):
            self.assertTrue(range_.kind == RangeKind.OPEN_RIGHT)
        elif isinstance(range_._lower, Included) and isinstance(range_._upper, Included):
            self.assertTrue(range_.kind == RangeKind.CLOSED)
        elif isinstance(range_._lower, Excluded) and isinstance(range_._upper, Excluded):
            self.assertTrue(range_.kind == RangeKind.OPEN)
        elif isinstance(range_._lower, Excluded) and isinstance(range_._upper, Included):
            self.assertTrue(range_.kind == RangeKind.OPEN_LEFT)

    @given(st.floats(allow_nan=False) | st.datetimes() | st.integers())
    def test_for_empty_range(self, point):
        a = point
        r1 = Range.new_open_right(a, a)
        r2 = Range.new_open_left(a, a)
        r3 = Range.new_open(a, a)
        r4 = Range.new_closed(a, a)
        self.assertEmpty(r1)
        self.assertEmpty(r2)
        self.assertEmpty(r3)
        self.assertNonEmpty(r4)

    @given(_sorted_points(many=3))
    def test_for_point_in_range(self, points):
        # any point in between the extremes will allways be in the range
        a, b, c = points
        self.assertIn(b, Range.new_open_right(a, c))
        self.assertIn(b, Range.new_open_left(a, c))
        self.assertIn(b, Range.new_open(a, c))
        self.assertIn(b, Range.new_closed(a, c))
        self.assertIn(b, Range.new_closed(b, b))

    @given(
        st.one_of(
            st.floats(max_value=10000.0, min_value=-10000.0),
            st.datetimes(),
            st.integers(max_value=10000, min_value=-10000),
        )
    )
    def test_bounds_operations(self, bound):
        assert ~Excluded(bound) == Included(bound)
        assert not (Excluded(Infinity) and Excluded(-Infinity))
        assert Included.safer_new(Infinity) == Excluded(Infinity)
        assert Included.safer_new(-Infinity) == Excluded(-Infinity)
        assert Included.safer_new(bound) == Included(bound)
        assert Included(bound)
        assert Excluded(bound)
        assert not (Excluded(-Infinity) and Excluded(-Infinity))
        assert Included(bound) != Excluded(bound)

    @given(
        st.floats() | st.datetimes() | st.integers(),
        st.sampled_from([Included, Excluded]),
    )
    def test_bounds_normalization(self, bound, Bound_):
        assert Bound.normalized_lower_bound(None, Bound_) == Excluded(-Infinity)
        assert Bound.normalized_upper_bound(None, Bound_) == Excluded(Infinity)

        assert Bound.normalized_lower_bound(bound, Excluded) == Excluded(bound)
        assert Bound.normalized_upper_bound(bound, Excluded) == Excluded(bound)

        assert Bound.normalized_lower_bound(bound, Included) == Included(bound)
        assert Bound.normalized_upper_bound(bound, Included) == Included(bound)


class EqSet(DomainCase):
    @given(equivalence_sets)
    def test_sample_returns_the_same(self, eqset):
        sample = eqset.sample
        # test sample returns the same object, and not an arbitrary one
        self.assertIs(sample, eqset.sample)
        self.assertIs(sample, eqset.sample)

    @given(equivalence_sets)
    def test_sample_is_in_the_set(self, eqset):
        sample = eqset.sample
        if eqset:
            self.assertIn(sample, eqset)
        else:
            self.assertIs(sample, None)

    @given(st.lists(equivalence_sets, min_size=1, max_size=20))
    def test_join_doesnt_mutate_args(self, eqsets):
        copies = [copy(r) for r in eqsets]
        EquivalenceSet.join(*eqsets)
        self.assertEqual(eqsets, copies)

    @given(equivalence_sets)
    def test_containment_of_the_empty_eqset(self, eqset):
        self.assertLessEqual(EquivalenceSet([]), eqset)
        self.assertGreaterEqual(eqset, EquivalenceSet([]))

    def _test_containment_of_union(self, eqset, other):
        self.assertLessEqual(eqset, eqset | other)
        self.assertLessEqual(other, eqset | other)
        self.assertSampleIsMember(eqset, eqset | other)
        self.assertSampleIsMember(other, eqset | other)

    @given(
        non_empty_equivalence_sets | integer_nonempty_intervalsets,
        non_empty_equivalence_sets,
    )
    def test_containment_of_union_other_eqset(self, other, eqset):
        self._test_containment_of_union(other, eqset)

    @given(
        non_empty_equivalence_sets | integer_nonempty_intervalsets,
        non_empty_equivalence_sets,
    )
    def test_containment_of_union_eqset_other(self, other, eqset):
        self._test_containment_of_union(eqset, other)

    def _test_containment_of_intersection(self, eqset, other):
        intersection = eqset & other
        self.assertLessEqual(intersection, eqset)
        self.assertLessEqual(intersection, other)
        if intersection:
            self.assertIn(intersection.sample, eqset)
            self.assertIn(intersection.sample, other)

    @given(equivalence_sets | integer_intervalsets, equivalence_sets)
    def test_containment_of_intersection_any_eqset(self, other, eqset):
        self._test_containment_of_intersection(other, eqset)

    @given(equivalence_sets | integer_intervalsets, equivalence_sets)
    def test_containment_of_intersection_eqset_any(self, other, eqset):
        self._test_containment_of_intersection(eqset, other)

    def _test_diff_containment(self, eqset, other):
        diff = eqset.diff(other)
        self.assertLessEqual(diff, eqset)

    @given(equivalence_sets | integer_intervalsets, equivalence_sets)
    def test_diff_containment_other_eqset(self, other, eqset):
        self._test_diff_containment(other, eqset)

    @given(equivalence_sets | integer_intervalsets, equivalence_sets)
    def test_diff_containment_eqset_other(self, other, eqset):
        self._test_diff_containment(eqset, other)


class IntervalTest(DomainCase):
    @given(
        st.lists(float_ranges(kinds=kinds_with_samples), max_size=5)
        | st.lists(date_ranges(kinds=kinds_with_samples), max_size=5)
    )
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_sample_is_member(self, ranges):
        interval = IntervalSet(ranges)
        self.assertSampleIsMember(interval, interval)

    @given(st.lists(float_ranges(), max_size=5) | st.lists(date_ranges(), max_size=5))
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_interval_filters_emptys(self, ranges):
        interval = IntervalSet(ranges)
        for range_ in interval.ranges:
            self.assertNonEmpty(range_)

    def _test_interval_diff_produces_disjoint_ranges(self, r1, r2):
        i1 = IntervalSet(r1)
        i2 = IntervalSet(r2)
        idiff = i1.diff(i2)
        assert not any(a & b for a, b in combinations(idiff.ranges, 2))

    @given(
        st.lists(float_ranges(), max_size=5),
        st.lists(float_ranges(), max_size=5),
    )
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_float_interval_diff_produces_disjoint_ranges(self, r1, r2):
        self._test_interval_diff_produces_disjoint_ranges(r1, r2)

    @given(st.lists(float_ranges(), min_size=2, max_size=5), st.data())
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_float_interval_diff_produces_disjoint_ranges_non_empty_diff(self, r1, data):
        last_range = r1[-1]
        next_range = _build_next_range(last_range, 1, data)
        r2 = r1 + [next_range]
        self._test_interval_diff_produces_disjoint_ranges(r1, r2)

    @given(
        st.lists(date_ranges(), min_size=2, max_size=5),
        st.lists(date_ranges(), max_size=5),
    )
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_date_interval_diff_produces_disjoint_ranges(self, r1, r2):
        self._test_interval_diff_produces_disjoint_ranges(r1, r2)

    def _test_interval_diff_produces_sorted_ranges(self, r1, r2):
        i1 = IntervalSet(r1)
        i2 = IntervalSet(r2)
        idiff = i1.diff(i2)
        assert idiff.ranges == tuple(sorted(idiff.ranges, key=operator.attrgetter("lowerbound")))

    @given(
        st.lists(float_ranges(), min_size=2, max_size=5),
        st.lists(float_ranges(), max_size=5),
    )
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_float_interval_diff_produces_sorted_ranges(self, r1, r2):
        self._test_interval_diff_produces_sorted_ranges(r1, r2)

    @given(
        st.lists(date_ranges(), min_size=2, max_size=5),
        st.lists(date_ranges(), max_size=5),
    )
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_date_interval_diff_produces_sorted_ranges(self, r1, r2):
        self._test_interval_diff_produces_sorted_ranges(r1, r2)

    @given(st.lists(date_ranges(), min_size=2, max_size=5), st.data())
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_date_interval_diff_produces_sorted_ranges_non_emptydiff(self, r1, data):
        last_range = r1[-1]
        next_range = _build_next_range(last_range, DAY, data)
        r2 = r1 + [next_range]
        self._test_interval_diff_produces_sorted_ranges(r1, r2)

    def _test_interval_diff_produces_non_toucheing_ranges(self, r1, r2):
        i1 = IntervalSet(r1)
        i2 = IntervalSet(r2)
        idiff = i1.diff(i2)
        assert not idiff.ranges or idiff.ranges == idiff.ranges[0].join(*idiff.ranges[1:])

    @settings(deadline=None, suppress_health_check=[HealthCheck.too_slow])
    @given(
        st.lists(float_ranges(), min_size=2, max_size=5),
        st.lists(float_ranges(), max_size=5),
    )
    def test_float_interval_diff_produces_non_toucheing_ranges(self, r1, r2):
        self._test_interval_diff_produces_non_toucheing_ranges(r1, r2)

    @settings(deadline=None, suppress_health_check=[HealthCheck.too_slow])
    @given(st.lists(float_ranges(), min_size=2, max_size=5), st.data())
    def test_float_interval_diff_produces_non_toucheing_ranges_non_emptydiff(self, r1, data):
        last_range = r1[-1]
        next_range = _build_next_range(last_range, 1, data)
        r2 = r1 + [next_range]
        self._test_interval_diff_produces_non_toucheing_ranges(r1, r2)

    @given(
        st.lists(date_ranges(), min_size=2, max_size=5),
        st.lists(date_ranges(), max_size=5),
    )
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_date_interval_diff_produces_non_toucheing_ranges(self, r1, r2):
        self._test_interval_diff_produces_non_toucheing_ranges(r1, r2)

    @given(
        st.lists(date_ranges(), max_size=5),
        st.lists(date_ranges(), max_size=5),
    )
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_containment_of_union(self, r1, r2):
        i1 = IntervalSet(r1)
        i2 = IntervalSet(r2)
        union = i1 | i2
        self.assertLessEqual(i1, union)
        self.assertLessEqual(i2, union)
        self.assertSampleIsMember(i1, union)
        self.assertSampleIsMember(i2, union)

    @given(
        st.lists(date_ranges(), max_size=5),
        st.lists(date_ranges(), max_size=5),
    )
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_containment_of_intersection(self, r1, r2):
        i1 = IntervalSet(r1)
        i2 = IntervalSet(r2)
        intersection = i1 & i2
        self.assertLessEqual(intersection, i1)
        self.assertLessEqual(intersection, i2)
        self.assertSampleIsMember(intersection, i1)
        self.assertSampleIsMember(intersection, i2)

    @given(
        st.lists(date_ranges(), max_size=5),
        st.lists(date_ranges(), max_size=5),
    )
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_containment_of_diff(self, r1, r2):
        i1 = IntervalSet(r1)
        i2 = IntervalSet(r2)
        diff = i1.diff(i2)
        self.assertLessEqual(diff, i1)

    @given(st.lists(date_ranges(), max_size=5))
    @settings(suppress_health_check=[HealthCheck.too_slow])
    def test_equal_intervals(self, r1):
        i1 = IntervalSet(r1)
        shuffle(r1)
        i2 = IntervalSet(r1)
        self.assertEqual(i1, i2)


@given(get_date_member_strategy())
def test_get_date_member_strategy_produces_nonimaginary_unfolded_dates(d):
    assert getattr(d, "fold", 0) == 0
    assert getattr(d, "tzinfo", None) is None


def _build_next_range(r, delta, data):
    Bound1, Bound2 = data.draw(st.tuples(type_of_bounds, type_of_bounds))
    lowerbound = r.upperbound + delta
    upperbound = lowerbound + delta
    return Range(Bound1(lowerbound), Bound2(upperbound))


# TODO: Move this to xoutil
def zip_map(funcs, iterable):
    for f, x in zip(funcs, iterable):
        yield f(x)



import os
from hypothesis import settings, Verbosity, HealthCheck

settings.register_profile("ci", max_examples=100)
settings.register_profile("dev", max_examples=50)
settings.register_profile("debug", max_examples=10, verbosity=Verbosity.verbose)

settings.register_profile(
    "coverage",
    max_examples=50,
    suppress_health_check=(HealthCheck.filter_too_much, HealthCheck.too_slow),
)

settings.load_profile(os.getenv("HYPOTHESIS_PROFILE", "dev"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from xotl.tools.objects import FinalSubclassEnumeration


def DynamicClassEnumeration(superclass):  # pragma: no cover
    return FinalSubclassEnumeration(superclass, dynamic=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# copyright (c) merchise autrement [~º/~] and contributors
# all rights reserved.
#
# this is free software; you can do what the licence file allows you to.
#
"""Implements trees for efficient searching."""

from __future__ import annotations

import operator
import typing as t
from dataclasses import dataclass, field
from itertools import takewhile

from xotl.tools.infinity import Infinity

from .ranges import Range, TOrd
from .tracing import get_module_sentry_spanner

T = t.TypeVar("T")


@dataclass(init=False, unsafe_hash=True)
class Cell(t.Generic[T, TOrd]):
    """A data-carrying open-right range."""

    # NB: Introducing any kind of Range here entails modifications to the
    # algorithms below.
    lowerbound: TOrd
    upperbound: TOrd
    data: T

    __slots__ = ("lowerbound", "upperbound", "data")

    def __init__(
        self,
        lower: t.Optional[TOrd],
        upper: t.Optional[TOrd],
        data: T,
    ) -> None:
        self.lowerbound = t.cast(TOrd, -Infinity if lower is None else lower)
        self.upperbound = t.cast(TOrd, Infinity if upper is None else upper)
        self.data = data

    @classmethod
    def from_bounds(
        cls,
        lower: t.Optional[TOrd],
        upper: t.Optional[TOrd],
        data: T,
    ) -> Cell[T, TOrd]:  # pragma: no cover
        """Return a cell from the boundaries of an open-right range."""
        return cls(lower, upper, data)

    @classmethod
    def from_range(cls, r: Range[TOrd], data: T) -> Cell[T, TOrd]:
        """Return a cell casting `r` to an open-right range.

        .. note:: We disregard the kind of range, cells are always open-right.

        """
        return cls(r.lowerbound, r.upperbound, data)

    def __contains__(self, which):  # pragma: no cover
        return self.lowerbound <= which < self.upperbound

    def __bool__(self) -> bool:  # pragma: no cover
        return self.lowerbound < self.upperbound


@dataclass(init=False, repr=False)
class IntervalTree(t.Generic[T, TOrd]):
    """A data-containing generic IntervalTree.

    Intended usage:

    - Create the tree by calling the classmethod `from_cells`:meth:.

      The `cells <Cell>`:class: which represent the boundaries of the interval
      related to an opaque data item.  The intervals are allowed over any type
      ``TOrd`` (i.e, for which `<=` is defined and total).

      The current implementation regards all intervals as open-right, closed-left
      (see `Cell.from_range`:meth:).

    - Use `get`:meth: or `__getitem__`:meth: (i.e ``tree[x]``) to get the
      intervals ``x`` belongs to.

    - Use ``x in tree`` to test if ``x`` belongs to at least one interval.

    - ``bool(tree)`` is True when the tree contains no cells.

    .. warning:: You SHOULDN'T mutate any of the internal attributes.  The API
       is just the class method `from_cells`:meth: and then getting the cells
       containing a point the with either `get`:meth: of `__getitem__`:meth:.

    .. seealso:: Section 10.1 of [deBerg2008]_.

    .. [deBerg2008] Mark de Berg, et al. *Computational Geometry. Algorithms
       and Applications. Third Edition*. Springer Verlag. 2008.

    """

    _impl: _ORIntervalTree[T, TOrd] = field(init=False)

    def __init__(self, cells: t.Sequence[Cell[T, TOrd]]) -> None:
        self._impl = _ORIntervalTree.from_cells(cells)

    @classmethod
    def from_cells(
        cls,
        cells: t.Sequence[Cell[T, TOrd]],
    ) -> IntervalTree[T, TOrd]:
        """Build an IntervalTree from the given sequence of cells."""
        return cls(cells)

    def __bool__(self):
        return bool(self._impl)

    def __contains__(self, query: TOrd):
        return query in self._impl

    def get(self, which: TOrd) -> t.Tuple[Cell[T, TOrd], ...]:
        """Return the sequence of cells to which `query` belongs.

        If `query` is not in any of the cells, return the empty tuple.

        """
        return self._impl.get(which)

    def __getitem__(self, which: TOrd) -> t.Tuple[Cell[T, TOrd], ...]:
        """Return the sequence of cells to which `query` belongs.

        If `query` is not in any of the cells, raise KeyError.

        """
        return self._impl[which]

    @property
    def depth(self):
        "The depth of the IntervalTree."
        return self._impl.depth


@dataclass
class _IntervalTreeImplementation(t.Generic[T, TOrd]):
    pass


@dataclass
class _ORIntervalTree(_IntervalTreeImplementation[T, TOrd]):
    "Implementation of IntervalTree with open-right cells."

    center: t.Optional[TOrd]
    cells: t.Sequence[Cell[T, TOrd]]  # Ordered by lowerbound
    cells_by_upper: t.Sequence[Cell[T, TOrd]]  # Ordered in reverse by upperbound.
    left: t.Optional[_ORIntervalTree]
    right: t.Optional[_ORIntervalTree]

    __slots__ = ("center", "cells", "cells_by_upper", "left", "right")

    @classmethod
    def from_cells(cls, cells: t.Sequence[Cell[T, TOrd]]) -> _ORIntervalTree[T, TOrd]:
        return cls._from_sorted_cells(sorted(tuple(cell for cell in cells if cell), key=_getlower))

    def __bool__(self) -> bool:
        return bool(self.cells)

    def __contains__(self, which: TOrd):
        return bool(self.get(which))

    @classmethod
    def _from_sorted_cells(
        cls,
        cells: t.Sequence[Cell[T, TOrd]],
    ) -> _ORIntervalTree[T, TOrd]:
        with sentry_span("IntervalTree._from_sorted_cells"):
            left_cells = []
            middle_cells = []
            right_cells = []
            if cells:
                middle = len(cells) // 2
                middle_cell = cells[middle]
                center: t.Optional[TOrd] = middle_cell.lowerbound
                for cell in cells:
                    if cell.upperbound <= center:
                        left_cells.append(cell)
                    elif center < cell.lowerbound:
                        right_cells.append(cell)
                    else:
                        assert center in cell
                        middle_cells.append(cell)
            else:
                center = None
            result = cls(
                center=center,
                cells=tuple(middle_cells),
                cells_by_upper=tuple(sorted(middle_cells, key=_getupper, reverse=True)),
                left=cls._from_sorted_cells(left_cells) if left_cells else None,
                right=cls._from_sorted_cells(right_cells) if right_cells else None,
            )
            return result

    def __getitem__(self, which: TOrd) -> t.Tuple[Cell[T, TOrd], ...]:
        """Return all cells that contains the given item."""
        result = self.get(which)
        if not result:
            raise KeyError(which)
        return result

    def get(self, query: TOrd) -> t.Tuple[Cell[T, TOrd], ...]:
        with sentry_span("IntervalTree.get"):
            return tuple(self.iter_get(query))

    def iter_get(self, query: TOrd) -> t.Iterator[Cell[T, TOrd]]:
        if not self:
            return
        if query <= self.center:
            # Since self.cells is ordered by lowerbound, only *the first*
            # cells for which the lowerbound is less or equal to the query
            # contain the queried value.  Looking any further won't find any
            # match.
            #
            # See page 222 of [deBerg2008].
            yield from takewhile(lambda c: c.lowerbound <= query, self.cells)
        else:
            # The same as before but with `cells_by_upper`.
            yield from takewhile(lambda c: query < c.upperbound, self.cells_by_upper)
        if query < self.center and self.left:
            yield from self.left.get(query)
        elif query > self.center and self.right:
            yield from self.right.get(query)

    @property
    def depth(self):
        "The depth of the IntervalTree."
        return 1 + max(
            left.depth if (left := self.left) is not None else 0,
            right.depth if (right := self.right) is not None else 0,
        )


_getlower = operator.attrgetter("lowerbound")
_getupper = operator.attrgetter("upperbound")

sentry_span = get_module_sentry_spanner(__name__)


if t.TYPE_CHECKING:
    from datetime import datetime

    Cell(1.0, 1.0, "Something")
    IntervalTree.from_cells([Cell(1.0, 1.0, "float"), Cell(1.0, 2.0, "float")])
    IntervalTree.from_cells([Cell(1, 1, "int"), Cell(-101, 21, "int")])
    IntervalTree.from_cells([
        Cell(datetime(2007, 3, 22), datetime(2022, 3, 22), "dt"),
        Cell(datetime(1978, 10, 21), datetime(2022, 10, 21), "dt"),
    ])

    IntervalTree.from_cells([Cell(1.0, 1.0, "float"), Cell(1, 2, "int")])  # type: ignore



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../"))

# -- Project information -----------------------------------------------------

project = "xotless"
author = "Merchise Autrement"

from datetime import datetime  # noqa

copyright = f"2012-{datetime.now().year} {author} [~º/~] and Contributors"
del datetime


# The full version, including alpha/beta/rc tags
try:
    from xotless._version import version

    release = version
except ImportError:
    from importlib import metadata
    from importlib.metadata import PackageNotFoundError

    try:
        dist = metadata.distribution("xotless")
        version = release = dist.version
    except PackageNotFoundError:
        version = "dev"
        release = "dev"


# -- General configuration ---------------------------------------------------

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
default_role = "code"
html_theme = "furo"
html_theme_options = {
    "light_css_variables": {
        "font-stack--monospace": '"Roboto Mono", "SFMono-Regular", Menlo, Consolas, Monaco, "Liberation Mono", "Lucida Console", monospace',
    },
}
html_static_path = ["_static"]
html_css_files = ["custom.css"]

# Configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "py3": ("https://docs.python.org/3.10/", None),
    "hypothesis": ("https://hypothesis.readthedocs.io/en/latest/", None),
    "xotl.tools": ("https://merchise-autrement.gitlab.io/xotl.tools/", None),
}

# Maintain the cache forever.
intersphinx_cache_limit = 365

autosummary_generate = True

# Produce output for todo and todolist
todo_include_todos = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import contextlib

from xotl.tools.context import context


def ReentrantContext(context_identifier):
    """Return a reentrant context manager.

    Entering a reentrant context the first time setups a dictionary that is
    shared with all nested entries to the same context.

    Examples:

       >>> CONTEXT1 = ReentrantContext(object())
       >>> with CONTEXT1() as d1:
       ...    with CONTEXT1() as d2:
       ...        assert d1 is d2

    However, non-nested entries to the context share no common dictionary.

    """

    @contextlib.contextmanager
    def reentrant_context():
        ctx = context[context_identifier]
        if ctx:
            yield ctx["memory"]
        else:
            with context(context_identifier, memory={}) as ctx:
                yield ctx["memory"]

    return reentrant_context
